/* Copyright (C) 2008, Cyril Hu (cyrilhu@gmail.com)

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA. */

/* ADT of queue (FIFO), in 3 forms.
 * DYNAMIC: slowest but flexible, handy when elem structure is really big.
 * STATIC: fastest but must know maximum size/len of queue.
 * BALANCED: general purpose circular queue, balanced performance/memory usage.
 * see http://en.wikipedia.org/wiki/Queue_(data_structure) for detail.
 */
#ifndef _QUEUE_H
#define _QUEUE_H

#include <stddef.h>
#include "ulbool.h"
#include "uldef.h"
_BEGIN_DECLS

#ifdef ADT_QUEUE_BALANCED
#define QUEUE_INCREMENT 64		/* memory reallocation increment. */
typedef struct {
	const void *data;		/* const pointer to external storage. */
} queue_node_t;				/* ADT of queue node. */
typedef struct {
	queue_node_t *front;		/* pointer to front node. */
	queue_node_t *rear;		/* pointer to rear node. */
	queue_node_t *memstart;		/* beginning of the allocated memory. */
	size_t memunits;		/* number of allocated memory units. */
} queue_t;				/* ADT of queue. */
#endif

#ifdef ADT_QUEUE_DYNAMIC
#include "sllist.h"
typedef sll_node_t queue_node_t;	/* reuse SLL node as queue node. */
typedef sll_t queue_t;			/* reuse SLL as ADT of queue. */
#endif

#ifdef ADT_QUEUE_STATIC
#define QUEUE_MAX_NODES 128		/* max nodes queue contains. */
typedef struct {
	const void *data;		/* const pointer to external storage. */
} queue_node_t;				/* ADT of queue node. */
typedef struct {
	queue_node_t e[QUEUE_MAX_NODES];/* data body, size unchangable. */
	queue_node_t *front;		/* pointer to front node. */
	queue_node_t *rear;		/* pointer to rear node. */
} queue_t;				/* ADT of queue. */
#endif

/* Init queue ADT, call this before any following operations. */
void queue_init(queue_t *p);

/* Attach a node to queue and make it new rear.
 * return false on malloc() failure, else true. */
bool queue_attach(queue_t *p, const void *data);

/* Detach current front from queue.
 * return external storage pointer or NULL on empty queue. */
const void *queue_detach(queue_t *p);

/* Peek specific node's element content,
 * return the pointer on valid offset, otherwise NULL. */
const void *queue_peek(queue_t *p, size_t offset);

/* Get number of node(s) queue contains. */
size_t queue_getlen(queue_t *p);

/* Whether queue is empty(no node) or not. */
bool queue_isempty(queue_t *p);

/* Clear queue content, making it empty/reusable. */
void queue_clear(queue_t *p);

/* Free allocated queue memory. */
void queue_destroy(queue_t *p);

_END_DECLS
#endif
