/* Copyright (C) 2008, Cyril Hu (cyrilhu@gmail.com)

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA. */

/* ADT of stack (LIFO), in 3 forms.
 * DYNAMIC: slowest but flexible, handy when elem structure is really big.
 * STATIC: fastest but must know maximum size/len of stack.
 * BALANCED: general purpose stack, balanced performance/memory usage.
 * see http://en.wikipedia.org/wiki/Stack_(data_structure) for detail.
 */
#ifndef _STACK_H
#define _STACK_H

#include <stddef.h>
#include "ulbool.h"
#include "uldef.h"
_BEGIN_DECLS

#ifdef ADT_STACK_BALANCED
#define STACK_INCREMENT 64		/* memory reallocation increment. */
typedef struct {
	const void *data;		/* const pointer to external storage. */
} stack_node_t;				/* ADT of stack node. */
typedef struct {
	stack_node_t *base;		/* pointer to base (bottom) node. */
	stack_node_t *top;		/* pointer to top (working end) node. */
	size_t memunits;		/* number of allocated memory units. */
} stack_t;				/* ADT of stack. */
#endif

#ifdef ADT_STACK_DYNAMIC
#include "sllist.h"
typedef sll_node_t stack_node_t;	/* reuse SLL node as stack node. */
typedef sll_t stack_t;			/* reuse SLL as ADT of stack. */
#endif

#ifdef ADT_STACK_STATIC
#define STACK_MAX_NODES 128		/* max nodes stack contains. */
typedef struct {
	const void *data;		/* const pointer to external storage. */
} stack_node_t;				/* ADT of stack node. */
typedef struct {
	stack_node_t e[STACK_MAX_NODES];/* data body, size unchangable. */
	stack_node_t *base;		/* pointer to base (bottom) node. */
	stack_node_t *top;		/* pointer to top (working end) node. */
} stack_t;				/* ADT of stack. */
#endif

/* Init stack ADT, call this before any following operations. */
void stack_init(stack_t *p);

/* Push a node into stack. return false on malloc() failure, else true. */
bool stack_push(stack_t *p, const void *data);

/* Popup top node from stack. 
 * return external storage pointer or NULL on empty stack. */
const void *stack_pop(stack_t *p);

/* Whether stack is empty(no node) or not. */
bool stack_isempty(stack_t *p);

/* Clear stack content, making it empty/reusable. */
void stack_clear(stack_t *p);

/* Peek specific node's element content,
 * return the pointer on valid offset, otherwise NULL. */
const void *stack_peek(stack_t *p, size_t offset);

/* Get number of node(s) stack contains. */
size_t stack_getlen(stack_t *p);

/* Free allocated stack memory. */
void stack_destroy(stack_t *p);

_END_DECLS
#endif
