/* Copyright (C) 2008, Cyril Hu (cyrilhu@gmail.com)

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA. */

#if defined(STR_DEBUG) || defined(STR_STRSSTR_DEBUG)
#include <stdio.h>
#endif
#include <stddef.h>
#include <string.h>
#include "ulbool.h"

/* Substitute all matching sub(s) with repl in src, 
 * src MUST be big enough to hold all changes!
 * return a pointer to changed src. if flag is true,
 * do it recursively (changed src becomes input again).
 *
 * slow (O(n square)) on large input, but generally 
 * OK and convenient when strlen(repl) <= strlen(sub). */
char *strsstr(char *src, const char *sub, const char *repl, bool flag)
{
	char *p, *t;
	size_t ls = strlen(sub), lr = strlen(repl);

#ifdef STR_DEBUG
	printf("\nstrsstr(): flag=%u\n", (unsigned)flag);
	printf("strsstr(): src=\"%s\"\n", src);
	printf("strsstr(): sub=\"%s\"\n", sub);
	printf("strsstr(): repl=\"%s\"\n", repl);
#endif
	/* sanity check. */
	if(!ls) {
#ifdef STR_DEBUG
		printf("strsstr(): strlen(sub)=0, sanity check failed.\n");
#endif
		goto done;
	}
	/* avoid possible infinite loop. */
	if(flag && strstr(repl, sub)) {
#ifdef STR_DEBUG
		printf("strsstr(): sub is a subset of repl, must avoid infinite loop.\n");
#endif
		flag = false;
	}

loop:	/* do it recursively . */
	if(!strstr(p=src,sub))
		goto done;
	if(ls > lr) { /* extra copy need. O(n square). */
#ifdef STR_DEBUG
		printf("strsstr(): strlen(sub) > strlen(repl)\n");
#endif
		while(*p)
			if( (t = strstr(p,sub)) != NULL) {
#ifdef STR_DEBUG
				printf("strsstr(): matching:    \"%s\"\n", t);
#endif
				memcpy(t, repl, lr);
#ifdef STR_DEBUG
				printf("strsstr(): after memcpy:\"%s\"\n", t);
#endif
				strcpy(t+lr, t+ls);
#ifdef STR_DEBUG
				printf("strsstr(): after moving:\"%s\"\n", t);
#endif
				p=t+lr;
			} else break;
	} else if (ls == lr) { /* yes, O(n) ! */
#ifdef STR_DEBUG
		printf("strsstr(): strlen(sub) == strlen(repl)\n");
#endif
		while(*p) 
			if( (t = strstr(p,sub)) != NULL) {
#ifdef STR_DEBUG
				printf("strsstr(): matching:    \"%s\"\n", t);
#endif
				memcpy(t, repl, lr);
#ifdef STR_DEBUG
				printf("strsstr(): after memcpy:\"%s\"\n", t);
#endif
				p=t+lr;
			} else break;
	} else { /* extra backward copy need. O(n square). */
#ifdef STR_DEBUG
		printf("strsstr(): strlen(sub) < strlen(repl)\n");
#endif
		while(*p)
			if( (t = strstr(p,sub)) != NULL) {
#ifdef STR_DEBUG
				printf("strsstr(): matching:    \"%s\"\n", t);
#endif
				memmove(t+lr, t+ls, strlen(t+ls)+1);
#ifdef STR_DEBUG
				printf("strsstr(): after moving:\"%s\"\n", t);
#endif
				memcpy(t, repl, lr);
#ifdef STR_DEBUG
				printf("strsstr(): after memcpy:\"%s\"\n", t);
#endif
				p=t+lr;
			} else break;
	}
	if(flag)
		goto loop;
done:
#ifdef STR_DEBUG
	printf("strsstr(): changed src=\"%s\"\n", src);
#endif
	return src;
}

#ifdef STR_STRSSTR_DEBUG
int main(void)
{
	char buf[512]="hi, this is  hcq0411, i like   programming, do you    like it?";
	strsstr(buf, " ", "   ", true);
	strsstr(buf, "  ", " ", false);
	strsstr(buf, "  ", " ", true);
	strsstr(buf, "", "", false);
	strsstr(buf, "is", "isisisisisisis", false);
	strsstr(buf, "isis", "is", false);
	strsstr(buf, "isis", "is", true);
	strsstr(buf, "", "ABCD", true);
	strsstr(buf, "ABCD", "", true);
	strsstr(buf, "like", "HATE!!!", true);
	strsstr(buf, "HATE!!!", "hate", false);
	strsstr(buf, "hate", "hate!!!", true);
	strsstr(buf, "hate!!!", "LIKE", false);
	strsstr(buf, "i", "", true);
	strcpy(buf, "11234567112999999999");
	strsstr(buf, "12", "212", false);
	return 0;
}
#endif
